from enum import Enum

class Parser:
	def __init__(self, name, address):
		self.name = name
		self.address = address
		self.bytes_used = {}
		self.bytes_used_from_hammer_stats = {} # TODO: once breakpoints vs detailed are exclusive, this should be removed
		self.apply_count = {} # Per arena
		self.successful_parse_count = {}
		self.failed_parse_count = {}

	def name_parser(self, name):
		self.name = name

	# TODO: The parser could be named in h_do_parse by deducing from the vtable
	def get_name_or_placeholder(self):
		if self.name is None:
			return "(Type resolution pending)"
		else:
			return self.name

	# 'arena' is expected to be int, and size probably should be too
	def add_mem_use(self, arena, size):
		if self.bytes_used.setdefault(arena, None) is None:
			self.bytes_used[arena] = size
		else:
			self.bytes_used[arena] += size

	# 'arena' is expected to be int
	def get_mem_use(self, arena=None):
		if arena is None:
			return self.bytes_used
		else:
			return self.bytes_used.setdefault(arena, 0)

	def __str__(self):
		return "(" + str(self.name) + ", " + hex(self.address) + ")"

	# Return the highest per-arena allocation count
	# TODO: disambiguate "parse state" and "arena", possibly tracking both
	def get_arenamax(self):
		res = 0
		if self.bytes_used:
			res = max(self.bytes_used.values())
		return res

	def get_arenasum(self):
		res = 0
		if self.bytes_used:
			res = sum(self.bytes_used.values())
		return res

	def increment_apply_count(self, arena):
		applycount = self.apply_count.get(arena, 0)
		applycount += 1
		self.apply_count[arena] = applycount

	def get_apply_count(self, arena):
		return self.apply_count.get(arena, 0)

	def increment_successful_parse_count(self, arena):
		s_count = self.successful_parse_count.get(arena, 0)
		s_count += 1
		self.successful_parse_count[arena] = s_count

	def get_successful_parse_count(self, arena):
		return self.successful_parse_count.get(arena, 0)

	def increment_failed_parse_count(self, arena):
		f_count = self.failed_parse_count.get(arena, 0)
		f_count += 1
		self.failed_parse_count[arena] = f_count

	def get_failed_parse_count(self, arena):
		return self.failed_parse_count.get(arena, 0)

class StackEvent(Enum):
        PUSH = 1
        POP = 2

class ParserStack:
	def __init__(self, parse_state, arena, top_level_parse):
		self.parse_state = parse_state
		self.parse_state_gdbval = parse_state
		self.arena = arena
		self.arena_gdbval = arena
		# self.parse_state and self.arena is expected to be an int by TopLevelParse and is ultimately used to index into the memory usage dict
		# parse_state is set by top_level_parse.first_h_do_parse_after_h_packrat_parse() to be an int
		# TODO: remove the redundancy and convert to int when indexing the memory dict? maybe a get_parse_state_address() function?
		self.top_level_parse = top_level_parse # To avoid depending on top_level_parse.py
		self.p_stack = []
		self.unclaimed_mem_use = 0
		self.partial_alloc_counts = []
		self.bytes_at_enter = []
		self.cumulative_byte_differences = []
		self.total_byte_differences = []
		self.bytes_since_prev_h_do_parse = []
		self.stack_events = []

	# New approach:
	# ParserStack tracks what "context" allocations happen in. A context is defined as a set of stack frames operating on the same parser
	# There are two routes out of a context: either returning from the last stack frame, to a different stack frame's context, or by applying a new parser
	# In the "apply" route, eventually the context will be returned to.
	# When leaving via returning to a previous context, the context is left for good (until the parser is applied again)

	# The previous idea was figuring out how much had been allocated at the point of popping the parser off the stack
	# The code below in pop() means to add together the bytes allocated after h_do_parse() was entered, and before h_do_parse() was left, subtracting allocations that happened in the contexts that were entered into in the meantime

	# On entering h_do_parse(), it checks the difference in arena_malloc_bytes, and attributes it to the parser whose context the code just left
	# This takes care of one "leg" of the context. let's call the value of arena_malloc_bytes here amb_1
	# Next time we revisit that context, it'll have been in the context of a "pop" operation, just as we're returning from h_do_parse
	# arena_malloc_bytes will have increased by the allocations since we checked aml_1 (importantly, including the allocations done after entering h_do_parse(), that we had already attributed during the next push()
	# Let's call the value of arena_malloc_bytes at the point of returning, amb_2
	# amb_2 = amb_1 + allocations we care about for the context + allocations in other contexts + allocations that happened after returning

	# The trivial case is when there is a HParser that doesn't apply any others
	# In that case, amb_2-amb_1 is exactly the number of bytes allocated in the context of the parser

	# For higher-order parsers, you have
	# amb_1 + other contexts' allocations + amb_2

	# The problem to solve, broadly, is to "flatten" out a series of push and pop operations into an 1D list
	# Also somewhat analogous to matching parentheses

	# The new approach involves taking samples of arena_malloc_bytes during both push() and pop() and storing them in the same list, perhaps as a tuple (push_or_pop, arena_malloc_bytes)
	# On a pop() operation, to compute the difference, the code needs to find the corresponding push(), subtract amb_2-amb_1, then from the result, recursively subtract the (amb_2-amb_1) differences of the push()/pop() pairs inbetween
	# The push/pop operations are implicitly guaranteed to be balanced so long as execution proceeds normally

	def push(self, parser):
		self.p_stack.append(parser)
		self.partial_alloc_counts.append(None)
		if self.top_level_parse.memory_stat_method == HammerMemoryStatisticsMethod.DETAILED_ARENA_STATS:
			#TODO: check that self.arena is same as gdb.selected_frame()'s arena
			self.bytes_at_enter.append(int(self.arena_gdbval['arena_malloc_bytes']))
			self.cumulative_byte_differences = []
			self.stack_events.append((StackEvent.PUSH, int(self.arena_gdbval['arena_malloc_bytes'])))
			prev = 0
			if len(self.bytes_at_enter) > 1:
				prev = self.bytes_at_enter[-2]
			thusfar = self.bytes_at_enter[-1] - prev
			print("Bytes allocated since last h_do_parse:", thusfar, ", parser:", str(parser)) # DEBUG
			#import code; code.interact(local=locals()) # DEBUG
			#breakpoint() # DEBUG
			#if thusfar < 0:
			#       import code; code.interact(local=locals()) #DEBUG
			#TODO: it's more convenient to commit allocation info to the stats when returning
			# But the user doesn't really care, so going this route requires keeping track of allocations that would be committed,
			# and presenting it a unified manner at the UI layer
			# Additionally, it would probably make more sense to store it at the ParserStack, storing (frame, bytes) pairs
			self.bytes_since_prev_h_do_parse.append(thusfar)

	def pop(self):
		self.partial_alloc_counts.pop()
		parser_obj = self.peek()

		if self.top_level_parse.memory_stat_method == HammerMemoryStatisticsMethod.DETAILED_ARENA_STATS:
		#if self.extended_arena_stats_available(): # DEBUG
			#import code; code.interact(local=locals()) # DEBUG
			newbytes = int(self.parse_state_gdbval['arena']['arena_malloc_bytes'])
			self.stack_events.append((StackEvent.POP, int(self.parse_state_gdbval['arena']['arena_malloc_bytes'])))
			# TODO: index expression with len() seems unnecessary
			self.cumulative_byte_differences.insert(0, newbytes - self.bytes_at_enter[-(len(self.cumulative_byte_differences)+1)]) # TODO: consistency. this is the third way of answering the question, "how deep are we in the stack"
			allocated_bytes = self.cumulative_byte_differences[-len(self.cumulative_byte_differences)] - self.cumulative_byte_differences[-(len(self.cumulative_byte_differences)-1)]  # NB: len(self.cumulative_byte_differences) changed since the last line
			self.total_byte_differences.insert(-len(self.cumulative_byte_differences), allocated_bytes)

			allocs = self.compute_parserstack_frame_allocs()
			#print("adding mem use: parser:", str(parser_obj), "arena:", hex(int(self.arena)), "bytes:", allocated_bytes) # DEBUG
			print("adding mem use (alternate): parser:", str(parser_obj), "arena:", hex(int(self.arena)), "bytes:", allocs) # DEBUG
			parser_obj.add_mem_use(int(self.arena), allocs)
			# TODO: allocations can be attributed to individual h_do_parse(parser) calls, thus the same per-parser per-arena stats tracking can be done

		return self.p_stack.pop()

	def peek(self):
		try:
			retval = self.p_stack[-1]
		except IndexError:
			retval = None
		return retval

	def set_state(self, state):
		self.parse_state = state
	# Shortcut for setting the name property of the parser on the top of stack
	# In terms of tracing, *most* calls to a parser look something like this with the packrat backend:
	# h_do_parse()
	#	parse_foo()
	#		perform_lowlevel_parse()

	# perform_lowlevel_parse() is called when the memo table at that position is not filled in yet.
	# it calls the corresponding parse_* virtual function via the vtable, but other than that does not have type information
	# it's probably possible to extract type information, by comparing vtable addresses, but that seems painful

	# parse_foo() is the parser's corresponding virtual function in the frontend, which does not have the equivalent of a "this" pointer

	# So what we do to keep track of parsers is incrementally filling in the details for both

	# h_do_parse() is the backend's "actually run the parser" function, but does not get called for some parsers
	# (apparently mostly it's for higher-order parsers)
	# also contains the decision logic about whether to call perform_lowlevel_parse()

	# possible scenarios:
	# h_do_parse()
	#	perform_lowlevel_parse()
	#		parse_foo()

	# h_do_parse()
	#	perform_lowlevel_parse()

	# h_do_parse()
	def name_top_parser(self, name):
		self.p_stack[-1].name_parser(name)

	def add_mem_use_each(self, size):
		for p in self.p_stack:
			p.bytes_used += size

	def add_mem_use_top(self, size):
		self.p_stack[-1].bytes_used += size

	def show_stack(self):
		print("stack would be printed here. Depth:", len(self.p_stack))
		#print([(p.get_name_or_placeholder(), hex(p.address)) for p in self.p_stack])

	def depth(self):
		return len(self.p_stack)

	def compute_parserstack_frame_allocs(self):
		stack_depth = 0
		stack_events = None
		for index, ev in enumerate(self.stack_events[::-1]):
			# Since this is meant to be called from ParserStack.pop(), the assumption here is that the first element will be a StackEvent.POP
			# So we can stop gathering memory allocations at the point stack_depth becomes 0 again: that'll be the point in the list,
			# where the parser we just popped from the stack had been pushed on it. Allocations prior to that are not relevant and are not considered
			if ev[0] == StackEvent.POP:
				stack_depth += 1
			elif ev[0] == StackEvent.PUSH:
				stack_depth -= 1
				if stack_depth == 0:
					stack_events = self.stack_events[-index-1:]
					# Possible optimization:
					# Get only self.stack_events[-1], self.stack_events[-2], self.stack_events[-index-1], self.stack_events[-index]
					print(stack_events) # DEBUG
					break

		return self.compute_outer_frame(stack_events)

	def compute_outer_frame(self, stack_slice):
		if len(stack_slice) == 2:
			return stack_slice[1][1] - stack_slice[0][1]
		# The code below assumes len(stack_slice) >= 4, TODO: ensure that len(stack_slice) is never 3
		else:
			# Remove the "outer" frame, call the function recursively on the "inner" frames
			#new_slice = stack_slice[1:-1]
			#return (stack_slice[-1][1] - stack_slice[0][1]) - self.compute_outer_frame(new_slice)
			# alternatively:
			return (stack_slice[-1][1] - stack_slice[-2][1]) + (stack_slice[1][1] - stack_slice[0][1])

	def compute_pushes_at_end(self):
		# Throws StopIteration if no more items in the iterator match the condition
		last_pop = next(index for index,ev in enumerate(self.stack_events[::-1]) if ev[0] == StackEvent.POP, None)
		if last_pop is None:
			# No "pop" events in the list, stack_events is all pushes
			# Subtract the bytes allocated before the first push happened, from the bytes allocated by the most recent h_do_parse()
			return self.stack_events[0][1] - self.stack_events[-1][1]

		return self.stack_events[0][1] - self.stack_events[-index][1]

		#for index, ev in enumerate(self.stack_events[::-1]):
		#	if ev[0] == StackEvent.POP:
		#		last_pop = index
